#
# Copyright 2018 Agilx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

APPTUIT_PY_TOKEN = "APPTUIT_API_TOKEN"
APPTUIT_PY_TAGS = "APPTUIT_TAGS"
DEPRECATED_APPTUIT_PY_TOKEN = "APPTUIT_PY_TOKEN"
DEPRECATED_APPTUIT_PY_TAGS = "APPTUIT_PY_TAGS"
__version__ = '2.4.1'

from apptuit import pyformance, timeseries
from .apptuit_client import Apptuit, DataPoint, ApptuitException, ApptuitSendException, \
    TimeSeriesName, TimeSeries

__all__ = ['Apptuit', 'DataPoint', 'ApptuitException', 'TimeSeriesName', 'TimeSeries',
           'pyformance', 'timeseries', 'ApptuitSendException', '__version__']
